#include "Image.h"
#include "LZW.h"

// zlib
#include "zlib/include/zlib.h"

#ifdef WIN32
// link the static lib (win32)
#pragma comment (lib, "zlib/lib/zdll.lib" )
#endif

#include <stdio.h> //fread

const unsigned int PIET_COLORS[] = {
	0xFFC0C0, 0xFFFFC0, 0xC0FFC0, 0xC0FFFF, 0xC0C0FF, 0xFFC0FF,
	0xFF0000, 0xFFFF00, 0x00FF00, 0x00FFFF, 0x0000FF, 0xFF00FF,
	0xC00000, 0xC0C000, 0x00C000, 0x00C0C0, 0x0000C0, 0xC000C0,
	0xFFFFFF, 0x000000
};

int readFile(const char* filename, CompressedImage& output){
	unsigned int totalSize = 0;
	struct Data{
		unsigned char* data;
		unsigned int size;
		Data* next;
	};
	Data* in = new Data;
	if (in == 0)
		return OUT_OF_MEMORY;
	in->data=0;
	in->size=0;
	in->next=0;
	Data* curIn = in;
	FILE* input = fopen(filename,"rb");
	if (input == 0){
		return FILE_NOT_FOUND;
	}
	unsigned int read = 0;
	int lastError = SUCCESS;
	do {
		curIn->data = new unsigned char[2048];
		if (curIn->data == 0){
			lastError=OUT_OF_MEMORY;
			break;
		}
		read = (unsigned int)fread(curIn->data,1,2048,input);
		curIn->size=read;
		totalSize+=read;
		curIn->next = new Data;
		if (curIn->next == 0){
			lastError=OUT_OF_MEMORY;
			break;
		}
		curIn = curIn->next;
		curIn->data=0;
		curIn->size=0;
		curIn->next=0;
	} while(read!=0);

	fclose(input);

	if (lastError==SUCCESS){
		output.image = new unsigned char[totalSize];
		if (output.image == 0){
			lastError=OUT_OF_MEMORY;
		}
	}

	if (lastError!=SUCCESS){
		// clean up heap!
		while(in != 0){
			if (in->data!=0){
				delete in->data;
			}
			curIn = in;
			in = in->next;
			delete curIn;
		}
		return lastError;
	}
	output.size = totalSize;
	// copy data from in to output.image. free heap while doing this.
	unsigned int j=0;
	while(in != 0){
		if (in->data!=0){
			for (unsigned int i=0;i<in->size;i++){
				output.image[j++] = in->data[i];
			}
			delete in->data;
		}
		curIn = in;
		in = in->next;
		delete curIn;
	}
	return SUCCESS;
}

int gif2bitmap(CompressedImage input, Bitmap& output){
	if (input.size < 6)
		return INVALID_INPUT_FORMAT;

	unsigned int position = 0;

	if (input.image[position++] != 'G')
		return INVALID_INPUT_FORMAT;
	if (input.image[position++] != 'I')
		return INVALID_INPUT_FORMAT;
	if (input.image[position++] != 'F')
		return INVALID_INPUT_FORMAT;
	if (input.image[position++] != '8')
		return INVALID_INPUT_FORMAT;
	if (input.image[position] != '7' && input.image[position] != '9')
		return INVALID_INPUT_FORMAT;
	position++;
	if (input.image[position++] != 'a')
		return INVALID_INPUT_FORMAT;

	if (input.size < 13)
		return IMAGE_CORRUPTED;

	unsigned short width = input.image[position++];
	width += ((unsigned short)(input.image[position++]))*256;
	unsigned short height = input.image[position++];
	height += ((unsigned short)(input.image[position++]))*256;

	bool hasGlobalColorMap = ((input.image[position] & 0x80)!=0);
	unsigned char bpp = (input.image[position++] & 0x7)+1;
	unsigned short colormapsize = (1<<bpp);

	unsigned char backgroundColor = input.image[position++];

	if (backgroundColor >= colormapsize)
		backgroundColor = 0;

	position++; //ignore pixel aspect ratio

	//Read global colormap
	unsigned int* globalColorMap = new unsigned int[colormapsize];
	if (globalColorMap == 0)
		return OUT_OF_MEMORY;
	for (unsigned short i=0;i<colormapsize;i++){
		globalColorMap[i] = 0;
	}
	if (hasGlobalColorMap){
		for (unsigned short i=0;i<colormapsize;i++){
			if (position+2 >= input.size){
				delete globalColorMap;
				return IMAGE_CORRUPTED;
			}
			globalColorMap[i] = input.image[position++];
			globalColorMap[i] += ((unsigned int)(input.image[position++]))*256;
			globalColorMap[i] += ((unsigned int)(input.image[position++]))*256*256;
		}
	}

	//create empty image (fill with background color)
	unsigned int* imageData = new unsigned int[height * width];
	if (imageData == 0) {
		delete globalColorMap;
		return OUT_OF_MEMORY;
	}
	for (unsigned int i=0;i<(unsigned int)height*(unsigned int)width;i++)
		imageData[i] = globalColorMap[backgroundColor];

	if (position >= input.size) {
		delete globalColorMap;
		delete imageData;
		return IMAGE_CORRUPTED;
	}

	for (unsigned char nextBlock=input.image[position++];nextBlock!=0x3B;nextBlock=input.image[position++]){
		switch (nextBlock){
			case 0x2C:
				{
					//IMAGE DATA
					unsigned short i_X;
					unsigned short i_Y;
					unsigned short i_Width;
					unsigned short i_Height;
					unsigned int readPos = 0;
					if (position+9 >= input.size) {
						delete globalColorMap;
						delete imageData;
						return IMAGE_CORRUPTED;
					}
					i_X = input.image[position++];
					i_X += ((unsigned short)(input.image[position++]))*256;
					i_Y = input.image[position++];
					i_Y += ((unsigned short)(input.image[position++]))*256;
					i_Width = input.image[position++];
					i_Width += ((unsigned short)(input.image[position++]))*256;
					i_Height = input.image[position++];
					i_Height += ((unsigned short)(input.image[position++]))*256;
					bool hasLocalColorMap = ((input.image[position] & 0x80)!=0);
					bool isInterlaced = ((input.image[position++] & 0x40)!=0);
					unsigned char localBpp;
					unsigned short localColormapsize;
					unsigned int* colorMap;
					if (hasLocalColorMap){
						localBpp = (input.image[position-1] & 0x7)+1;
						localColormapsize = (1<<localBpp);
						colorMap = new unsigned int[localColormapsize];
						if (colorMap == 0) {
							delete globalColorMap;
							delete imageData;
							return OUT_OF_MEMORY;
						}
						for (int i=0;i<localColormapsize;i++){
							if (position+2 >= input.size){
								delete globalColorMap;
								delete imageData;
								delete colorMap;
								return IMAGE_CORRUPTED;
							}
							colorMap[i] = input.image[position++];
							colorMap[i] += ((unsigned int)(input.image[position++]))*256;
							colorMap[i] += ((unsigned int)(input.image[position++]))*256*256;
						}
					}else{
						localBpp = bpp;
						localColormapsize = colormapsize;
						colorMap = globalColorMap;
					}
					//read LZW-compressed image data
					if (position >= input.size){
						delete globalColorMap;
						delete imageData;
						if (hasLocalColorMap)
							delete colorMap;
						return IMAGE_CORRUPTED;
					}
					LZW lzw(input.image[position++]);

					//now read image data...
					if (position >= input.size){
						delete globalColorMap;
						delete imageData;
						if (hasLocalColorMap)
							delete colorMap;
						return IMAGE_CORRUPTED;
					}
					for (unsigned char dataBlockSize=input.image[position++];dataBlockSize!=0;dataBlockSize=input.image[position++]){

						char dataBlock[255];
						for (unsigned char i=0;i<dataBlockSize;i++){
							if (position >= input.size){
								break; //TODO: return IMAGE CORRUPTED: UNEXPECTED END OF FILE
							}
							dataBlock[i] = input.image[position++];
						}
						for (unsigned char i=0;i<dataBlockSize;i++){
							char* decompressedData=0;
							unsigned short decompressedDataLength=0;
							if (lzw.isFinished()){
								// ERROR: Data in GIF image block after end of LZW data.
								//TODO: return IMAGE CORRUPTED
							}
							lzw.decodeByte(dataBlock[i],decompressedData,decompressedDataLength); //TODO: out_of_memory innerhalb der decodeByte-Methode zurückgeben/abfangen!!!
							if (decompressedData != 0){
								for (unsigned short j=0;j<decompressedDataLength;j++){
									unsigned int pixelValue = 0;
									if (decompressedData[j] < localColormapsize)
										pixelValue = colorMap[decompressedData[j]];
									//Write pixelValue to imageData-field
									if (readPos < ((unsigned int)i_Width)*((unsigned int)i_Height))
										GIF_writeData(pixelValue,readPos++,i_X,i_Y,i_Width,i_Height,imageData,width,isInterlaced);
									else if (readPos++ == i_Width*i_Height){
										// Too many data in GIF image block.
										//TODO: return IMAGE CORRUPTED
									}
								}
								delete decompressedData;
							}
						}
						if (position >= input.size){
							break; //TODO: return IMAGE CORRUPTED: UNEXPECTED END OF FILE
						}
					}
					if (readPos < ((unsigned int)i_Width * (unsigned int)i_Height)) {
						// Too less data in GIF image block.
						//TODO: return IMAGE CORRUPTED
					}
					if (!lzw.isFinished()){
						// LZW data in GIF image block not terminated.
						// TODO: return IMAGE CORRUPTED
					}
					if (hasLocalColorMap)
						delete colorMap;
				}
				break;
			case 0x21:
				{
					//EXTENSION BLOCK - IGNORE! - TODO: fread durch lesen aus input.image ersetzen
					position++; //extension type (ignore it)
					unsigned char bs = input.image[position++]; //block size
					if (position >= input.size){
						break; //TODO: return IMAGE CORRUPTED unexpected end of file.
					}
					while (bs != 0){
						position+=bs;
						if (position >= input.size)
							break; //TODO: return IMAGE CORRUPTED unexpected end of file.
						bs = input.image[position++]; //block size
					}
				}
				break;
			default:
				//ERROR!
				delete imageData;
				delete globalColorMap;
				return IMAGE_CORRUPTED;
		}
		if (position >= input.size)
			break; //TODO: unexpected END OF FILE
	}
	delete globalColorMap;

	output.size[0] = width;
	output.size[1] = height;
	output.image = imageData;

	return SUCCESS;
}

void GIF_writeData(unsigned int pixelValue, unsigned int readPos, unsigned int i_X, unsigned int i_Y, unsigned int i_Width, unsigned int i_Height, unsigned int* imageData, unsigned int width, bool interlaced){
	unsigned int y;
	if (interlaced){
		if ((readPos / i_Width)*8<i_Height){
			y = (readPos / i_Width)*8;
		}else if (((readPos / i_Width) - ((i_Height+7)/8))*8+4<i_Height){
			y = ((readPos / i_Width) - ((i_Height+7)/8))*8+4;
		}else if (((readPos / i_Width) - ((i_Height+7)/8) - ((i_Height+3)/8))*4+2 < i_Height){
			y = ((readPos / i_Width) - ((i_Height+7)/8) - ((i_Height+3)/8))*4+2;
		}else{
			y = ((readPos / i_Width) - ((i_Height+7)/8) - ((i_Height+3)/8)- ((i_Height+1)/4) )*2+1;
		}
	}else{
		y = (readPos / i_Width)+i_Y;
	}
	imageData[((readPos % i_Width)+i_X) + y*width] = ((pixelValue%256)*256*256 + ((pixelValue/256)%256)*256 + (pixelValue/(256*256)));
}

int png2bitmap(CompressedImage input, Bitmap& output){
	if (input.size < 8)
		return INVALID_INPUT_FORMAT;

	unsigned int position = 0;
	
	if (input.image[position++] != 0x89)
		return INVALID_INPUT_FORMAT;
	if (input.image[position++] != 'P')
		return INVALID_INPUT_FORMAT;
	if (input.image[position++] != 'N')
		return INVALID_INPUT_FORMAT;
	if (input.image[position++] != 'G')
		return INVALID_INPUT_FORMAT;
	if (input.image[position++] != 0x0D)
		return INVALID_INPUT_FORMAT;
	if (input.image[position++] != 0x0A)
		return INVALID_INPUT_FORMAT;
	if (input.image[position++] != 0x1A)
		return INVALID_INPUT_FORMAT;
	if (input.image[position++] != 0x0A)
		return INVALID_INPUT_FORMAT;

	//NOW critical chunks: IHDR, then PLTE and IDAT, IDAT finished by IEND.
	//IGNORE ancillary  chunks. Break up on unknown critical chunks.

	unsigned int width = 0;
	unsigned int height = 0;
	unsigned char* imageData = 0;
	unsigned int imageDataSize = 0;
	typedef struct ChunkList {
		unsigned char* chunkData;
		unsigned int chunkDataSize;
		ChunkList* next;
	} ChunkList;
	ChunkList* chunklist = new ChunkList;
	if (chunklist == 0)
		return OUT_OF_MEMORY;
	chunklist->chunkData = 0;
	chunklist->chunkDataSize = 0;
	ChunkList* lastChunk = chunklist;
	unsigned char bitdepth = 0;
	unsigned char colortype = 0;
	bool interlaced = false;
	unsigned int* colormap = 0;
	unsigned int colormapSize = 0;

	unsigned int chunkType;
	unsigned int chunkLen;

	if (position+7 >= input.size){
		delete chunklist;
		return IMAGE_CORRUPTED;
	}

	chunkLen = ((unsigned int)(input.image[position++]))*256*256*256;
	chunkLen += ((unsigned int)(input.image[position++]))*256*256;
	chunkLen += ((unsigned int)(input.image[position++]))*256;
	chunkLen += ((unsigned int)(input.image[position++]));


	chunkType = ((unsigned int)(input.image[position++]));
	chunkType += ((unsigned int)(input.image[position++]))*256;
	chunkType += ((unsigned int)(input.image[position++]))*256*256;
	chunkType += ((unsigned int)(input.image[position++]))*256*256*256;

	if (chunkType != 0x52444849 || chunkLen!=0x0D){
		delete chunklist;
		return IMAGE_CORRUPTED;
	}
	//read IHDR-chunk
	if (position+12 >= input.size){
		delete chunklist;
		return IMAGE_CORRUPTED;
	}
	width = ((unsigned int)(input.image[position++]))*256*256*256;
	width += ((unsigned int)(input.image[position++]))*256*256;
	width += ((unsigned int)(input.image[position++]))*256;
	width += ((unsigned int)(input.image[position++]));


	height = ((unsigned int)(input.image[position++]))*256*256*256;
	height += ((unsigned int)(input.image[position++]))*256*256;
	height += ((unsigned int)(input.image[position++]))*256;
	height += ((unsigned int)(input.image[position++]));

	bitdepth = input.image[position++];
	colortype = input.image[position++];

	unsigned char CIn;
	CIn = input.image[position++];
	if (CIn != 0x00){
		delete chunklist;
		return IMAGE_CORRUPTED;
	}

	CIn = input.image[position++];
	if (CIn != 0x00){
		delete chunklist;
		return IMAGE_CORRUPTED;
	}

	CIn = input.image[position++];
	if (CIn != 0x00 && CIn != 0x01){
		delete chunklist;
		return IMAGE_CORRUPTED;
	}
	interlaced = (CIn != 0x00);

	//now CRC, ignore it
	position+=4;

	if (colortype != 0 && colortype != 2 && colortype != 3 && colortype != 4 && colortype != 6){
		delete chunklist;
		return IMAGE_CORRUPTED;
	}
	if (bitdepth != 8){
		switch (colortype) {
			case 0:
				if (bitdepth != 1 && bitdepth != 2 && bitdepth != 4 && bitdepth != 16){
					delete chunklist;
					return IMAGE_CORRUPTED;
				}
				break;
			case 2:
			case 4:
			case 6:
				if (bitdepth != 16){
					delete chunklist;
					return IMAGE_CORRUPTED;
				}
				break;
			case 3:
				if (bitdepth != 1 && bitdepth != 2 && bitdepth != 4){
					delete chunklist;
					return IMAGE_CORRUPTED;
				}
				break;
			default:
				delete chunklist;
				return IMAGE_CORRUPTED;
		}
	}

	if (colortype == 0 || colortype == 4){
		//greyscale image. cannot be a senseful Piet program.
		delete chunklist;
		return NOT_YET_IMPLEMENTED_FEATURES_DETECTED;
	}

	//now read chunks:
	while(1){
		if (position >= input.size){
			break; //END OF FILE
		}
		if (position+7 >= input.size){
			delete chunklist; //UNEXPECTED END OF FILE
			return IMAGE_CORRUPTED;
		}

		chunkLen = ((unsigned int)(input.image[position++]))*256*256*256;
		chunkLen += ((unsigned int)(input.image[position++]))*256*256;
		chunkLen += ((unsigned int)(input.image[position++]))*256;
		chunkLen += ((unsigned int)(input.image[position++]));


		chunkType = ((unsigned int)(input.image[position++]));
		chunkType += ((unsigned int)(input.image[position++]))*256;
		chunkType += ((unsigned int)(input.image[position++]))*256*256;
		chunkType += ((unsigned int)(input.image[position++]))*256*256*256;

		switch (chunkType){
			case 0x45544C50:
				//PLTE
				if (colortype == 0 || colortype == 4){
					//error: PLTE not allowed in grayscale image
					//TODO: clean up heap
					delete chunklist;
					return IMAGE_CORRUPTED;
				}
				if (colortype != 3){
					//PLTE is optional. ignore it.
					position += chunkLen;
					if (position > input.size){
						//error: unexpected eof
						//TODO: clean up heap
						delete chunklist; //UNEXPECTED END OF FILE
						return IMAGE_CORRUPTED;
					}
					break;
				}
				if (colormap != 0){
					//error: PLTE already exists
					//TODO: clean up heap
					delete chunklist;
					delete colormap;
					return IMAGE_CORRUPTED;
				}
				{
					// read data and store them to colormap!
					colormapSize = (1<<bitdepth);
					if (colormapSize > 256)
						colormapSize = 256;
					colormap = new unsigned int[colormapSize];
					if (colormap == 0){
							// TODO: clean up heap
							delete chunklist;
							return OUT_OF_MEMORY;
					}
					for (unsigned int i=0;i<colormapSize;i++)
						colormap[i] = 0;
					for (unsigned int i=0;i<colormapSize && i<chunkLen/3;i++){
						if (position+2 >= input.size){
							//error: unexpected end of file
							//TODO: clean up heap
							delete chunklist; //UNEXPECTED END OF FILE
							delete colormap;
							return IMAGE_CORRUPTED;
						}
						colormap[i] = ((unsigned int)(input.image[position++]))*256*256;
						colormap[i] += ((unsigned int)(input.image[position++]))*256;
						colormap[i] += ((unsigned int)(input.image[position++]));
					}
				}
				break;
			case 0x54414449:
				//IDAT
				if (imageData != 0 || chunklist == 0 || lastChunk == 0){
					//error: IDAT after IEND
					//TODO: clean up heap
					delete colormap;
					return IMAGE_CORRUPTED;
				}
				if (chunkLen > 0){
					if (position+chunkLen > input.size){
						//error: unexpexcted end of file
						//TODO: clean up heap
						delete colormap;
						return IMAGE_CORRUPTED;
					}
					//add chunk to chunk list
					ChunkList* newChunk = new ChunkList;
					if (newChunk == 0){
						//TODO: clean up heap
						delete chunklist;
						return OUT_OF_MEMORY;
					}
					newChunk->next = 0;
					newChunk->chunkDataSize = chunkLen;
					newChunk->chunkData = new unsigned char[chunkLen];
					if (newChunk->chunkData == 0){
						//TODO: clean up heap
						delete chunklist;
						return OUT_OF_MEMORY;
					}
					for (unsigned int i=0;i<chunkLen;i++){
						newChunk->chunkData[i] = input.image[position++];
					}
					lastChunk->next = newChunk;
					lastChunk = newChunk;
				}
				break;
			case 0x444E4549:
				//IEND
				if (imageData != 0 || imageDataSize != 0){
					//error: IEND after IEND
					//ignore...
					position += chunkLen;
					if (position > input.size){
						//error: unexpected eof
						//TODO: clean up heap
						delete chunklist; //UNEXPECTED END OF FILE
						return IMAGE_CORRUPTED;
					}
					break;
				}
				if (chunklist == lastChunk || chunklist == 0 || lastChunk == 0){
					//error: no IDATs before IEND
					//TODO: clean up heap
					delete chunklist; //UNEXPECTED END OF FILE
					return IMAGE_CORRUPTED;
				}
				{
					ChunkList* iterator = chunklist;
					while (iterator != 0){
						imageDataSize += iterator->chunkDataSize;
						iterator = iterator->next;
					}
					imageData = new unsigned char[imageDataSize];
					iterator = chunklist;
					//chunklist = 0;
					//lastChunk = 0;
					unsigned int imageDataPos = 0;
					while (iterator != 0){
						if (iterator->chunkData != 0){
							for (unsigned int i=0;i<iterator->chunkDataSize;i++){
								if (imageDataPos >= imageDataSize){
									//ERROR: unexpected error
									//TODO: clean up heap
									delete chunklist;
									return IMAGE_CORRUPTED;
								}
								imageData[imageDataPos++] = iterator->chunkData[i];
							}
						}
						if (iterator->chunkDataSize > 0 && iterator->chunkData != 0)
							delete iterator->chunkData;
						iterator = iterator->next;
						delete chunklist;
						chunklist = iterator;
					}
					if (imageDataPos < imageDataSize){
						// Inconsistent data. TODO: return IMAGE_CORRUPTED
					}
					imageDataSize = imageDataPos;

					position += chunkLen;
					if (position > input.size){
						//error: unexpected eof
						//TODO: clean up heap
						delete chunklist; //UNEXPECTED END OF FILE
						return IMAGE_CORRUPTED;
					}
				}
				break;
			default:
				if ((chunkType & 0x20)==0){
					//unknown critical chunk
					//TODO: clean up heap
					delete chunklist;
					return NOT_YET_IMPLEMENTED_FEATURES_DETECTED;
				}
				position += chunkLen;
				if (position > input.size){
					//error: unexpected eof
					//TODO: clean up heap
					delete chunklist;
					return IMAGE_CORRUPTED;
				}
				break;
		}
		//now CRC, ignore it
		position += 4;
		if (position > input.size){
			//unexpected EOF.
			//TODO: clean up heap.
			delete chunklist;
			return IMAGE_CORRUPTED;
		}
	}

	unsigned char valuesPesPixel = ((colortype & 1)==1?1:(1+((colortype&4)>>2)+(colortype&2)));
	unsigned char passes = (interlaced?7:1);
	unsigned int* lineBytes = new unsigned int[passes];
	unsigned int* lineWidth = new unsigned int[passes];
	unsigned int* lineNumbers = new unsigned int[passes];
	unsigned long maxImDataSize = 0;
	if (!interlaced){
		lineBytes[0] = (bitdepth*valuesPesPixel*width+7)/8;
		lineWidth[0] = width;
		lineNumbers[0] = height;
	}else{
		lineBytes[0] = (bitdepth*valuesPesPixel*((width+7)/8)+7)/8;
		lineWidth[0] = (width+7)/8;
		lineBytes[1] = (bitdepth*valuesPesPixel*((width+3)/8)+7)/8;
		lineWidth[1] = (width+3)/8;
		lineBytes[2] = (bitdepth*valuesPesPixel*((width+3)/4)+7)/8;
		lineWidth[2] = (width+3)/4;
		lineBytes[3] = (bitdepth*valuesPesPixel*((width+1)/4)+7)/8;
		lineWidth[3] = (width+1)/4;
		lineBytes[4] = (bitdepth*valuesPesPixel*((width+1)/2)+7)/8;
		lineWidth[4] = (width+1)/2;
		lineBytes[5] = (bitdepth*valuesPesPixel*((width)/2)+7)/8;
		lineWidth[5] = (width)/2;
		lineBytes[6] = (bitdepth*valuesPesPixel*width+7)/8;
		lineWidth[6] = width;
		lineNumbers[0] = (height+7)/8;
		lineNumbers[1] = (height+7)/8;
		lineNumbers[2] = (height+3)/8;
		lineNumbers[3] = (height+3)/4;
		lineNumbers[4] = (height+1)/4;
		lineNumbers[5] = (height+1)/2;
		lineNumbers[6] = height/2;
	}
	for (unsigned char i=0;i<(passes);i++)
		maxImDataSize += lineNumbers[i]*(lineBytes[i]+1);
	
	unsigned char* decompressedImageData = new unsigned char[maxImDataSize];
	if (decompressedImageData == 0){
		return OUT_OF_MEMORY;
	}

	int result = uncompress(decompressedImageData, &maxImDataSize, imageData, imageDataSize);
	if (result != 0){
		// Error: gzip could not decompress PNG data.
		//TODO: clean up heap
		return GZIP_ERROR;
	}

	unsigned int offset = 0;
	for (unsigned char pass = 0;pass<passes;pass++){
	
		for (unsigned int i=0;i<lineNumbers[pass];i++){
			//unfilter line
			unsigned char filterType = decompressedImageData[i*(lineBytes[pass]+1)+offset];
			if (filterType == 0)
				continue;
			for (unsigned int j=0;j<lineBytes[pass];j++){
				int lineOffset = i*(lineBytes[pass]+1);
				int position = lineOffset+1+j;
				int position_a = 0;
				if (bitdepth < 8)
					position_a = position - 1;
				else
					position_a = position - (bitdepth*valuesPesPixel)/8;
				int position_b = position - (lineBytes[pass]+1);
				int position_c = position_a - (lineBytes[pass]+1);
				signed int a;
				signed int b;
				signed int c;
				if (position_a > lineOffset){
					a = decompressedImageData[position_a+offset];
					if (position_b > 0){
						c = decompressedImageData[position_c+offset];
					}else{
						c = 0;
					}
				}else{
					a=0;
					c=0;
				}
				if (position_b > 0){
					b = decompressedImageData[position_b+offset];
				}else{
					b = 0;
				}
				switch (filterType){
					case 1:
						decompressedImageData[position+offset] += ((unsigned char)a);
						break;
				case 2:
					if (position_b > 0)
						decompressedImageData[position+offset] += ((unsigned char)b);
					break;
				case 3:
					if (position_a > lineOffset && position_b > 0)
						decompressedImageData[position+offset] += (unsigned char)((a+b)/2);
					break;
				case 4:
					{
						signed int p = a+b-c;
						signed int pa = (p-a>0?p-a:a-p);
						signed int pb = (p-b>0?p-b:b-p);
						signed int pc = (p-c>0?p-c:c-p);
						decompressedImageData[position+offset] += ((unsigned char)((pa <= pb && pa <= pc)?a:(pb <= pc?b:c)));
					}
					break;
				default:
					//ERROR
					//Unsupported filter type in PNG image.
					//TODO: clean up heap
					return NOT_YET_IMPLEMENTED_FEATURES_DETECTED;
				}
			}
		}
		offset += lineNumbers[pass]*(lineBytes[pass]+1);
	}

	//decompressedImageData now contains all unfiltered image lines. first byte of every line must be ignored.
	unsigned int* imageD = new unsigned int[width*height];
	if (imageD == 0){
		return OUT_OF_MEMORY;
	}
	unsigned char* data = new unsigned char[valuesPesPixel];
	if (data == 0){
		delete imageD;
		return OUT_OF_MEMORY;
	}
	offset = 0;
	unsigned int xpos[] = {0,4,0,2,0,1,0};
	unsigned int ypos[] = {0,0,4,0,2,0,1};
	unsigned int xrepeat[] = {8,8,4,4,2,2,1};
	unsigned int yrepeat[] = {8,8,8,4,4,2,2};
	if (!interlaced){
		xrepeat[0] = 1;
		yrepeat[0] = 1;
	}
	for (unsigned char pass = 0;pass<passes;pass++){
	
		for (unsigned int y=0;y<lineNumbers[pass];y++)
		for (unsigned int x=0;x<lineWidth[pass];x++){
			unsigned int noninterlacedPixelPos = x*xrepeat[pass]+xpos[pass] + (y*yrepeat[pass]+ypos[pass])*width;
			for (unsigned char v=0;v<valuesPesPixel;v++){
				data[v] = decompressedImageData[offset+y*(lineBytes[pass]+1)+1+((x*valuesPesPixel+v)*bitdepth)/8];
				if (bitdepth < 8){
					data[v] = (((unsigned char)(data[v]<<((x*valuesPesPixel+v)*bitdepth % 8)))>>(8-bitdepth));
				}
			}
			if (colortype==3){
				imageD[noninterlacedPixelPos] = colormap[data[0]];
			}else if (colortype==2 || colortype == 6){
				imageD[noninterlacedPixelPos] = ((unsigned int)(data[0]))*256*256+((unsigned int)(data[1]))*256+((unsigned int)(data[2]));
			}else if (colortype==0 || colortype == 4){
				imageD[noninterlacedPixelPos] = ((unsigned int)(data[0]))*256*256+((unsigned int)(data[0]))*256+((unsigned int)(data[0]));
			}else{
				//should not occur
				// Program flow error while reading PNG file.
				//TODO: clean up heap
				delete imageD;
				return INTERNAL_ERROR;
			}
		}
		offset += lineNumbers[pass]*(lineBytes[pass]+1);
	}

	delete lineBytes;
	delete lineNumbers;
	delete decompressedImageData;
	delete data;


	output.size[0] = width;
	output.size[1] = height;
	output.image = imageD;
	return SUCCESS;
}

int bmp2bitmap(CompressedImage input, Bitmap& output){

	unsigned int* imageData = 0;
	unsigned int width = 0;
	int height = 0;

	if (input.size < 2)
		return INVALID_INPUT_FORMAT;

	unsigned int position = 0;
	
	if (input.image[position++] != 'B')
		return INVALID_INPUT_FORMAT;
	if (input.image[position++] != 'M')
		return INVALID_INPUT_FORMAT;
	if (input.size < 18)
		return IMAGE_CORRUPTED;

	position += 8; // TODO what is ignored here..?
	
	unsigned int imDatOffset = ((unsigned int)(input.image[position++]));
	imDatOffset += ((unsigned int)(input.image[position++]))*256;
	imDatOffset += ((unsigned int)(input.image[position++]))*256*256;
	imDatOffset += ((unsigned int)(input.image[position++]))*256*256*256;

	unsigned int headerSize = ((unsigned int)(input.image[position++]));
	headerSize += ((unsigned int)(input.image[position++]))*256;
	headerSize += ((unsigned int)(input.image[position++]))*256*256;
	headerSize += ((unsigned int)(input.image[position++]))*256*256*256;
	
	if (headerSize < 40)
		return NOT_YET_IMPLEMENTED_FEATURES_DETECTED; //TODO: accept smaller BITMAPINFOHEADERs? (for older versions of bmp-files?)
	if (position+headerSize > input.size)
		return IMAGE_CORRUPTED; //unexpected eof
	width = ((unsigned int)(input.image[position++]));
	width += ((unsigned int)(input.image[position++]))*256;
	width += ((unsigned int)(input.image[position++]))*256*256;
	width += ((unsigned int)(input.image[position++]))*256*256*256;
	height = (int)((unsigned int)(input.image[position++]));
	height += (int)(((unsigned int)(input.image[position++]))*256);
	height += (int)(((unsigned int)(input.image[position++]))*256*256);
	height += (int)(((unsigned int)(input.image[position++]))*256*256*256);

	// ignore 2 bytes
	position += 2;

	unsigned int colordepth = 0;
	
	colordepth = ((unsigned int)(input.image[position++]));
	colordepth += ((unsigned int)(input.image[position++]))*256;

	if (colordepth != 24 && colordepth != 32)
		return NOT_YET_IMPLEMENTED_FEATURES_DETECTED; //TODO: support 16, 8 and 4 bits/pixel
	unsigned int compression = ((unsigned int)(input.image[position++]));
	compression += ((unsigned int)(input.image[position++]))*256;
	compression += ((unsigned int)(input.image[position++]))*256*256;
	compression += ((unsigned int)(input.image[position++]))*256*256*256;
	if (compression != 0)
		return NOT_YET_IMPLEMENTED_FEATURES_DETECTED; //TODO: support RLE-compression and BI_BITFIELDs...
	// ignore 4 bytes: size of image data in bytes TODO: do not ignore.
	// ignore 8 bytes. TODO: find out what is ignored here!
	position += 12;
	unsigned int sizeOfColormap = ((unsigned int)(input.image[position++]));
	sizeOfColormap += ((unsigned int)(input.image[position++]))*256;
	sizeOfColormap += ((unsigned int)(input.image[position++]))*256*256;
	sizeOfColormap += ((unsigned int)(input.image[position++]))*256*256*256;
	if (sizeOfColormap != 0)
		return NOT_YET_IMPLEMENTED_FEATURES_DETECTED; //TODO: support color maps...
	// ignore 4 bytes.
	position += 4;

	// END OF HEADER REACHED

	// IGNORE DATA BETWEEN HEADER AND DATAOFFSET
	position = imDatOffset;
	if (position > input.size) {
		// Unexpected end of bitmap file.
		return IMAGE_CORRUPTED;
	}

	bool startAtLastLine = true;
	unsigned int curY = (unsigned int)(height-1);
	if (height <= 0){
		height = -height;
		startAtLastLine = false;
		curY = 0;
	}

	imageData = new unsigned int[height * width];
	if (imageData == 0) {
		return OUT_OF_MEMORY;
	}

	int zeilenRest = (4-(((width%4) * (colordepth/8))%4))%4;

	for (unsigned int y=0;y<(unsigned int)height;y++){
		for (unsigned int x=0;x<width;x++){
			unsigned int IIn = 0;
			unsigned int multiplicator=1;
			for (unsigned int i=0;i<colordepth/8;i++){
				if (position >= input.size) {
					// Unexpected end of bitmap file.
					delete imageData;
					return IMAGE_CORRUPTED;
				}

				IIn += ((unsigned int)(input.image[position++]))*multiplicator;
				multiplicator *= 256;
			}
			imageData[curY*width+x] = (IIn & 0x00FFFFFF);
		}
		if (zeilenRest > 0)
			position += zeilenRest;
		if (startAtLastLine)
			curY--;
		else
			curY++;
	}

	if (position > input.size) {
		// Unexpected end of bitmap file.
		delete imageData;
		return IMAGE_CORRUPTED;
	}

	output.size[0] = width;
	output.size[1] = height;
	output.image = imageData;
	return SUCCESS;
}

int ppm2bitmap(CompressedImage input, Bitmap& output){ //TODO: binary PPM doesnt work correctly. E.g. piet_bfi.ppm
	unsigned int* imageData = 0;
	unsigned int position = 0;
	
	if (input.size < 2)
		return INVALID_INPUT_FORMAT;

	if (input.image[position++] != 'P')
		return INVALID_INPUT_FORMAT;
	if (input.image[position] != '3' && input.image[position] != '6')
		return INVALID_INPUT_FORMAT;

	bool binaer = (input.image[position++] == '6');

	unsigned int width;
	unsigned int height;
	unsigned int maxval;

	if (!PPM_readAsci(width,input,position))
		return IMAGE_CORRUPTED;
	if (!PPM_readAsci(height,input,position))
		return IMAGE_CORRUPTED;
	if (!PPM_readAsci(maxval,input,position))
		return IMAGE_CORRUPTED;

	if (binaer && maxval > 0xFFFF)
		return IMAGE_CORRUPTED;
	
	imageData = new unsigned int[height * width];
	if (imageData == 0)
		return OUT_OF_MEMORY;

	unsigned const C0 = ((192*maxval)/255); //C0 => ???
	for (unsigned int y=0;y<height;y++){
		for (unsigned int x=0;x<width;x++){
			unsigned int curcol = 0;
			for (int rgb=0;rgb<3;rgb++){

				unsigned int IIn = 0;

				if (binaer){
					if (maxval > 255){
						if (position >= input.size){
							delete imageData;
							return IMAGE_CORRUPTED;
						}
						IIn = input.image[position++]*256;
					}
					if (position >= input.size){
						delete imageData;
						return IMAGE_CORRUPTED;
					}
					IIn += input.image[position++]*256;
				}else{
					if (!PPM_readAsci(IIn,input,position)){
						delete imageData;
						return IMAGE_CORRUPTED;
					}
				}
				
				if (IIn >= maxval)
					IIn = 255;
				else if (IIn == 0)
					IIn = 0;
				else if (IIn == C0 || IIn+1 == C0 || IIn-1 == C0)
					IIn = 0xC0;
				else
					IIn = (IIn * 255) / maxval;

				curcol += (IIn << ((2-rgb)*8));
				
				imageData[y*width+x] = curcol;
			}
		}
	}
	output.image = imageData;
	output.size[0] = width;
	output.size[1] = height;

	return SUCCESS;
}

bool PPM_readAsci(unsigned int& value, CompressedImage in, unsigned int& position){

	if (position >= in.size){
		return false;
	}

	char input;
	do{
		do{
			if (position >= in.size){
				return false;
			}
			input = (char)in.image[position++];
		}while(input == ' ' || input == '\r' || input == '\n' || input == '\t' || input == 0x0b || input == 0);
		if (input == '#')
			do{
				if (position >= in.size){
					return false;
				}
				input = (char)in.image[position++];
			}while(input != '\r' && input != '\n');
		else
			break;
	}while(1);

	if (input < '0' || input > '9')
		return false;

	unsigned long temp = input-'0';
	while(1){
		if (position >= in.size){
			return true;
		}
		input = (char)in.image[position++];
		if (input == ' ' || input == '\r' || input == '\n' || input == '\t' || input == 0x0b || input == 0){
			value = (unsigned int)temp;
			return true;
		}
		if (input == '#'){
			value = (unsigned int)temp;
			//read until end of line or end of file
			do{
				if (position >= in.size){
					return true;
				}
				input = (char)in.image[position++];
			}while(input != '\r' && input != '\n');
			return true;
		}
		if (input < '0' || input > '9')
			return false;
		temp *= 10;
		temp += input-'0';

		if (temp > 0xFFFFFFFFl)
			return false;
	}
}

int compressed2bitmap(CompressedImage input, Bitmap& output){
	int result = gif2bitmap(input, output);
	if (result != INVALID_INPUT_FORMAT)
		return result;
	result = png2bitmap(input, output);
	if (result != INVALID_INPUT_FORMAT)
		return result;
	result = bmp2bitmap(input, output);
	if (result != INVALID_INPUT_FORMAT)
		return result;
	result = ppm2bitmap(input, output);
	return result;
}


unsigned int getCodelSize(Bitmap input, bool unknownPixelsAsWhite){
	unsigned int width = input.size[0];
	unsigned int height = input.size[1];
	unsigned int maxCs = (width < height?width:height);
	if (maxCs==0)
		return 0; //invalid image

	bool* divisors = new bool[maxCs+1];
	if (divisors == 0)
		return 0; //out of memory
	divisors[0]=false;
	divisors[1]=true;
	for (unsigned int i=2;i<maxCs+1;i++){
		divisors[i] = (width%i==0 && height%i==0);
	}

	unsigned int codelsize = 1;

	// start with divisor >1 and check if it could be a valid codel size.
	for (unsigned int i=2;i<maxCs+1;i++){
		if (divisors[i]){
			if (isValidCodelSize(i, input, unknownPixelsAsWhite)){
				// max codelsize is a multiple of found codel size.
				codelsize = i;
				for (unsigned int j=i+1;j<maxCs+1;j++){
					if (j%i!=0)
						divisors[j] = false;
				}
			}else{
				// max codelsize cannot be a multiple of not-matching codelsize.
				for (unsigned int j=i+1;j<maxCs+1;j++){
					if (j%i==0)
						divisors[j] = false;
				}
			}
		}
	}
	delete divisors;
	
	return codelsize;
}

bool isValidCodelSize(unsigned int codelsize, Bitmap input, bool unknownPixelsAsWhite){
	unsigned int width = input.size[0];
	unsigned int height = input.size[1];
	if (codelsize < 1)
		return false;
	if (codelsize==1)
		return true;
	if (codelsize > width || codelsize > height)
		return false;
	if (width % codelsize != 0 || height % codelsize != 0)
		return false;
	for (unsigned int y=0;y<height;y+=codelsize)
		for (unsigned int x=0;x<width;x+=codelsize){
			unsigned char color = getPietColor(input.image[x+y*width], unknownPixelsAsWhite); //nicht-piet-farben alle gleich betrachten!
			for (unsigned int cy = 0; cy < codelsize; cy++)
				for (unsigned int cx = 0; cx < codelsize; cx++)
					if (getPietColor(input.image[(x+cx)+(y+cy)*width], unknownPixelsAsWhite) != color)
						return false;
		}

	return true;
}

unsigned char getPietColor(unsigned int rgb, bool unknownPixelsAsWhite){
	unsigned char pc = (unknownPixelsAsWhite?WHITE:BLACK);
	for (unsigned char j=0;j<20;j++){
		if (rgb == PIET_COLORS[(int)j]){
			pc = j;
			break;
		}
	}
	return pc;
}

int bitmap2pietimage(Bitmap input, PietImage& output, unsigned int codelSize, bool unknownPixelsAsWhite){
	if (codelSize==0)
		codelSize=getCodelSize(input, unknownPixelsAsWhite);
	if (codelSize==0)
		codelSize=1;
	unsigned int pietWidth = input.size[0]/codelSize+(input.size[0]%codelSize!=0?1:0);
	unsigned int pietHeight = input.size[1]/codelSize+(input.size[1]%codelSize!=0?1:0);
	output.image = new unsigned char[pietWidth*pietHeight];
	if (output.image == 0)
		return OUT_OF_MEMORY;
	output.size[0] = pietWidth;
	output.size[1] = pietHeight;
	unsigned int pietY=0;
	for (unsigned int y=0;y<input.size[1];y+=codelSize){
		unsigned int pietX=0;
		for (unsigned int x=0;x<input.size[0];x+=codelSize){
			output.image[pietY*pietWidth+pietX] = getPietColor(input.image[y*input.size[0]+x], unknownPixelsAsWhite);
			pietX++;
		}
		pietY++;
	}
	output.size[0]=pietWidth;
	output.size[1]=pietHeight;
	return SUCCESS;
}

