#ifndef PIETIMAGE2PROGRAM_INCLUDED
#define PIETIMAGE2PROGRAM_INCLUDED

#include "Image.h"
#include "Program.h"

/**
 * The CodelCompileStack stores reachable Codels that might not have been processed yet.
 * At the beginning this stack will be initialized with the top left codel.
 * While processing a codel from the stack every reachable codel will be added if it has not been processed before.
 * The program has been converted completely if this stack is empty.
 * 
 * Some codels may be stored twice or more often so before processing a codel it should be checked if it has been processed already.
 * If any other codel of the same region has been processed already processing won't be neccessary.
 */
typedef struct CodelCompileStack {
	/**
	 * Position of the codel in the PietImage.
	 */
	unsigned int codel[2];
	/**
	 * Program flow direction (DP & CC) that should be processed.
	 */
	unsigned char d;
	/**
	 * Previous element in stack.
	 */
	CodelCompileStack* prev;
} CodelCompileStack;

/**
 * The PixelStack structure stores pixels that should be added to the current treated region.
 */
typedef struct PixelStack{
	unsigned int position[2];
	PixelStack* previous;
} PixelStack;

/**
 * Data structure to save pointers to PragramCommandList* that has to be modified if the corresponding command is added to a region.
 */
typedef struct ListOfProgramCommandList {
	ListOfProgramCommandList* next;
	ProgramCommandList** pcl;
} ListOfProgramCommandList;

/**
 * Represents a region of connected codels with the same color.
 * The exitpoints for each possible direction are stored as a position of the PietImage.
 * 
 * If a codel is to be processed via the CodelCompileStack its connected codels and position
 * of exit points has only to be searched once by saving them in this struct.
 * The exit point may be a black codel that cannot be entered. The real exit point and exit direction
 * must be processed and saved in the command array.
 * 
 */
typedef struct {
	//size of the region in codels (for push command)
	unsigned int size;
	//color of the region (0-19)
	unsigned char color;
	//exitpoints[DC/CC][x/y-value]: exit point inside the region
	//idea: each white codel gets an own region entry, containing the leftmost, rightmost, upmost and downmost pixel in straight.
	//then: if the codel next to the exitpoint is black the next codel should be searched using the new region at the exit point.
	//all visited points (coordinates and the CC/DP) have to be stored. if a visited point is reached again with same CC/DP the program terminates.
	//size of black and white regions doesn't matter.
	unsigned int exitpoints[8][2];
	// stores a pointer to the corresponding piet command of the program for each exit-direction. NULL if the command has not been generated yet.
	ProgramCommandList* command[8];

	// TODO: SPEICHERE FUER JEDE DER 8 MOEGLICHEN RICHTUNGEN EINE LISTE, DIE POINTER AUF ADRESSEN ENTHAELT, AN DENEN EIN ProgramCommandList*-Pointer steht,
	// DER AUF command[RICHTUNG] GESETZT WERDEN MUSS, SOBALD command[RICHTUNG] EXISTIERT!
	ListOfProgramCommandList* comeFrom[8];

	// id of the region. (not used)
	// unsigned int id;
} Region;
// Generierter Programmcode könnte weiter verkleinert werden, wenn 

/**
 * Creates a region containing the codel given and stores the region to the regionmap.
 *
 * MAYBE TODO: If a region for the codel given already exists simply return the region.
 *
 * @return A Pointer to the generated region.
 */
Region* createRegion(unsigned int codel[2], PietImage image, Region** regionmap);

/**
 * Reads in a PietImage struct and generates a Program struct.
 * A PietImage struct contains exactly one codel per pixel and color values from 0 to 19.
 * The Program struct is a useful representation of a program for easy building program code in any language.
 */
int pietimage2program(PietImage input, Program& output, bool spaceAfterNumOut=true, bool printQuestionMarkOnIn=false);

/**
 * Gets the coordinates of the next pixel in the specified direction.
 * Direction 0: right
 * Direction 1: down
 * Direction 2: left
 * Direction 3: up
 */
void getNeightbourPixel(unsigned int position[2], unsigned char direction, unsigned int neightbour[2]);

#endif

